#include <boost/any.hpp>
#include <iostream>
#include <string>
#include <stdexcept>
#ifdef WIN32
#include "windows.h"
#else
#include <dlfcn.h>
#endif

#include "../common.h"

#ifdef WIN32
 typedef HMODULE tLibHandle;
#else
 typedef void* tLibHandle;
#endif

using namespace std;
using namespace boost;

tLibHandle getLib(std::string libname){
#ifdef WIN32
	std::string real_lib_name = libname + ".dll";
	cout << "trying to open "<< real_lib_name <<endl;;
	HMODULE h = LoadLibrary(real_lib_name.c_str());
	if(!h){
		cerr << "Could not open lib1" <<endl;
		return 0;
	} 
	return h;
#else
#ifdef __APPLE__
	std::string real_lib_name = "lib" + libname + ".dylib";
	cout << "trying to open "<< real_lib_name <<endl;;
#else
	std::string real_lib_name = "lib" + libname + ".so";
	cout << "trying to open "<< real_lib_name <<endl;;
#endif
	void * h = dlopen(real_lib_name.c_str(), RTLD_LAZY|RTLD_GLOBAL);
	if(!h){
		throw std::runtime_error("Could not open " + libname);
	} 
	return h;
#endif
}

void* getFunction(tLibHandle h,  std::string funcname){
#ifdef WIN32
	void * pFunc = GetProcAddress(h, funcname.c_str());
#else
	void * pFunc = dlsym(h, funcname.c_str());
#endif
	if(!pFunc){
		throw std::runtime_error("Could not find " + funcname);
	}
	return pFunc;
}
/*
	SuperClass* (*pFunc2)() = (SuperClass* (*)())dlsym(h, "getSuperButIsDerived");
	if(!pFunc){
		cerr << "Could not find getSuperButIsDerived" <<endl;
		return;
	}	
	anys_holder a;
	try{
		
		cout << "Adress of typeid from main of anys_holder" << &typeid(anys_holder) <<endl;
		pFunc(a);
		cout << "Any casting ..... " << any_cast<SuperClass>(a.anys["lib1"]).m_name <<  "  ..... works" << endl;
	} catch( boost::bad_any_cast& bac){
		cout << "Ouch no any_casts :( " << bac.what() <<endl;
	}
	cout << "----------------dynamic_casting----------------" <<endl;

	cout << "Adress of typeid from main of Derived" << &typeid(Derived) <<endl;
	SuperClass* p = pFunc2();
	if(dynamic_cast<Derived*>(p)){
		cout << "Yeha. dynamic_cast works!" << endl;
	} else {
		cout << "Ouch, no dynamic casting!" << endl;
	}
	return a;*/


int main(){
	std::cout << "Adress of typeid from main of Derived " << &typeid(Derived) <<std::endl;
	std::cout << "Adress of typeid from main of SuperClass " << &typeid(SuperClass) <<std::endl;

	tLibHandle lib2 = getLib("anytestlib2");
	{
		SuperClass* (*pFunc)() = (SuperClass* (*)())getFunction(lib2, "getSuperButIsDerived");
		SuperClass* p = pFunc();
		if(dynamic_cast<Derived*>(p)){
			cout << "Yeha. dynamic_casting in main works!" << endl;
		} else {
			cout << "Ouch, no dynamic casting in main !" << endl;
		}
	}
	{
		void (*pFunc)(SuperClass*) = (void (*)(SuperClass*))getFunction(lib2, "extractDerivedFromSuper");
		pFunc(new Derived());
	}
	tLibHandle lib1 = getLib("anytestlib1");
	{
		SuperClass* (*pFunc)() = (SuperClass* (*)())getFunction(lib1, "getSuperButIsDerived");
		SuperClass* p = pFunc();

		void (*pFuncLib2)(SuperClass*) = (void (*)(SuperClass*))getFunction(lib2, "extractDerivedFromSuper");
		pFuncLib2(p);
		
	}
	
}
